;; -*- mode: emacs-lisp; lexical-binding: t; -*-

;;; files

;; (add-to-list 'backup-directory-alist (cons "roth_2015" "backups/"))

(defconst gi/roth/basedir "~/2020/05/201504-roth_2015/")
(defconst gi/roth/roth (concat gi/roth/basedir "roth_2015.roth"))
(defconst gi/roth/glossary (concat gi/roth/basedir "roth_2015.glossary"))
(defconst gi/roth/rglossary (concat gi/roth/basedir "roth_2015.rglossary"))
(defconst gi/roth/traduzione (concat gi/roth/basedir "roth_2015.traduzione"))
(defvar gi/roth/errors ()) 

;;; tables

(defconst gi/roth/elt-tables 
  '(("gi/roth/elt" . 1900)		; (chk elt)	-> len
    ("gi/roth/elt-type" . 1900)		; (chk elt)	-> str
    ("gi/roth/lne-text" . 4200)  	; (chk elt lne) -> str
    ("gi/roth/lne-pageskip" . 250)	; (chk elt lne) -> ((point . num) ...)
    ("gi/roth/lne-italic" . 300)        ; (chk elt lne) -> ((start . end) ...)
    ("gi/roth/lne-notesref" . 200)	; (chk elt lne) -> ((point . num) ...)
    ("gi/roth/lne-indexref" . 800)	; (chk elt lne) -> ((point . num) ...)
    ("gi/roth/lne-link" . 15)	        ; (chk elt lne) -> ((point . url))
    ))

(defconst gi/roth/elt-ttables 
  '(("gi/roth/elt-pageskip" . 250)	; (chk elt)	-> ((point . num) ...)
    ("gi/roth/elt-italic" . 300)        ; (chk elt)	-> ((start . end) ...)
    ("gi/roth/elt-bold" . 10)		; (chk elt)	-> ((start . end) ...)
    ("gi/roth/elt-notesref" . 200)	; (chk elt)	-> ((point . num) ...)
    ("gi/roth/elt-indexref" . 800)	; (chk elt)	-> ((point . num) ...)
    ("gi/roth/elt-link" . 15)	        ; (chk elt)	-> ((point . url) ...)
    ("gi/roth/lne" . 4200)		; (chk elt lne) -> (start . end)
    ))

(defconst gi/roth/gls-tables
  '(("gi/roth/gls" . 2200)              ; rstr ->
					; (((chk elt lne strt end) strt end) ...)
    ("gi/roth/gls-index" . 200)		; (chk elt) ->
					; (((chk elt lne strt end) strt end) ...)
    ("gi/roth/gls-names" . 400)		; str ->
					; (((chk elt lne strt end) strt end) ...)
    ))

(defconst gi/roth/gls-rtables
  '(("gi/roth/gls-r" . 3100)		; (chk elt lne) -> (rstr ...)
    ("gi/roth/gls-index-r" . 700) 	; (chk elt lne) -> ((chk elt) ...)
    ("gi/roth/gls-names-r" . 700) 	; (chk elt lne) -> (str ...)
    ))

(defconst gi/roth/trd-tables 
  '(("gi/roth/trd-italic" . 300)        ; (chk elt lne) -> ((start . end) ...)
    ("gi/roth/trd-notesref" . 200)	; (chk elt lne) -> ((point . num) ...)
    ("gi/roth/trd-indexref" . 800)	; (chk elt lne) -> ((point . num) ...)
    ("gi/roth/trd-link" . 15)	        ; (chk elt lne) -> ((point . url))
    ("gi/roth/trd-note" . 15)		; (chk elt lne) -> ((point . note))
    ("gi/roth/trd-text" . 4200)		; (chk elt lne) -> str
    ))

;;;; xhtml

(defun gi/roth/xhtml ()
  (let ((order '("front02.xhtml" "front03.xhtml"
		 "c01.xhtml" "c02.xhtml" "c03.xhtml" "c04.xhtml" "c05.xhtml"
		 "c06.xhtml" "c07.xhtml" "c08.xhtml" "c09.xhtml" "c10.xhtml"
		 "c11.xhtml" "c12.xhtml" "c13.xhtml" "c14.xhtml" "c15.xhtml"
		 "c16.xhtml"
		 "back01.xhtml" "back02.xhtml" "back03.xhtml")))
    (dotimes (x (length order))
      (with-temp-buffer
	(insert-file-contents (concat gi/roth/basedir "xhtml/" (nth x order)))
	(goto-char (point-min)) (while (re-search-forward ". . ." nil t)
				  (replace-match "[dots]"))
	(goto-char (point-min)) (while (re-search-forward " " nil t)
				  (replace-match " "))
	(gi/roth/sxhtml
	 (cdr (assoc 'body (libxml-parse-html-region (point-min) (point-max))))
	 x)))
    (maphash (lambda (k v) (puthash (list (car k) 1) "parte" gi/roth/elt-type))
	     gi/roth/elt-bold)
    (maphash
     (lambda (k v) (mapcar
	       (lambda (x) (mapcar
		       (lambda (y) (let ((z (gethash x gi/roth/lne)))
				(when (and (>= (car y) (car z))
					   (< (cdr y) (cdr z)))
				  (puthash x (cons (cons (- (car y) (car z))
							 (- (cdr y) (car z)))
						   (gethash
						    x gi/roth/lne-italic))
					   gi/roth/lne-italic))))
		       v))
	       (gi/roth/elt-lines k)))
     gi/roth/elt-italic)
    (cl-mapc
     (lambda (a b)
       (maphash
	(lambda (k v) (mapcar
		  (lambda (x) (mapcar
			  (lambda (y) (let ((z (gethash x gi/roth/lne)))
				   (when (and (>= (car y) (car z))
					      (<= (car y) (cdr z)))
				     (puthash x (cons
						 (cons (- (car y) (car z))
						       (cdr y))
						 (gethash x b))
					      b))))
			  v))
		  (gi/roth/elt-lines k)))
	a))
     (list gi/roth/elt-pageskip gi/roth/elt-notesref
	   gi/roth/elt-indexref gi/roth/elt-link)
     (list gi/roth/lne-pageskip gi/roth/lne-notesref
	   gi/roth/lne-indexref gi/roth/lne-link))))

(defun gi/roth/sxhtml (lst chk-idx)
  (let ((elt-idx 0))
    (dolist (x lst)
      (let* ((elt (list chk-idx elt-idx))
	     (a (cond ((= 18 chk-idx) (gi/roth/sxhtml-match-notes x))
		      ((= 19 chk-idx) (gi/roth/sxhtml-match-index x))
		      (t (gi/roth/sxhtml-match x))))
	     (cat (nth 0 a)) (key (nth 1 a)) (txt (nth 2 a)))
	(pcase cat
	  (:n nil)
	  (:a (puthash elt (cons (cons 0 txt) (gethash elt key)) key))
	  (:b (progn (puthash elt key gi/roth/elt-type)
		     (gi/roth/sxhtml-elt txt elt)
		     (setq elt-idx (1+ elt-idx))))
	  (:c  (dolist (z txt) ; non recursiveness bites here
		 (pcase z (`(li ,_ (p . (((class . "left ")) . ,y)))
			   (progn (puthash elt (concat "elemento " key)
					   gi/roth/elt-type)
				  (gi/roth/sxhtml-elt y elt)
				  (setq elt-idx (1+ elt-idx))
				  (setq elt (list chk-idx elt-idx))))
			(_ (push (list :list z) gi/roth/errors))))))))))

(defun gi/roth/sxhtml-match (x)
  (pcase x
    ((pred null) (list :n))
    ((pred (lambda (z) (and (stringp z) (string-match-p "^[ \n\t]*$" z))))
     (list :n))
    (`(div . (((class . "timebreak ")) . ,_)) (list :n))
    (`(a . (((id . ,y)))) (list :a gi/roth/elt-pageskip y))
    (`(p . (((class . "center sink")) . ,y)) (list :b "extra" y))
    (`(p . (((class . "left ")) . ,y)) (list :b "paragrafo" y))
    (`(p . (((class . "follow")) . ,y)) (list :b "paragrafo" y))
    (`(p . (((class . "center ")) ,_ (b nil ,y) ,_))
     (if (= (length y) 1) (list :b "extra" (list y))
       (list :b "sezione" (list y))))
    (`(blockquote ,_ ,_ (p . (((class . "left ")) . ,y)) ,_)
     (list :b "citazione" y))
    (`(blockquote ,_ ,_ (p . (((class . "right ")) . ,y)) ,_)
     (list :b "attribuzione" y))
    (`(blockquote ,_ ,_ (p . (((class . "follow")) . ,y)) ,_)
     (list :b "citazione" y))
    (`(h1 . (((class . "fronttitle ")) . ,y)) (list :b "titolo" y))
    (`(h1 . (((class . "chaptertitle ")) . ,y)) (list :b "capitolo" y))
    (`(h1 . (((class . "backtitle ")) . ,y)) (list :b "titolo" y))
    (`(ol . ,y) (list :c "lista ordinata" y))
    (`(ul . ,y) (list :c "lista non ordinata" y))
    (_ (push (list :1 x) gi/roth/errors))))

(defun gi/roth/sxhtml-match-notes (x)
  (pcase x
    ((pred null) (list :n))
    ((pred (lambda (z) (and (stringp z) (string-match-p "^[ \n\t]*$" z))))
     (list :n))
    (`(div . (((class . "timebreak ")) . ,_)) (list :n))
    (`(a . (((id . ,y)))) (list :a gi/roth/elt-pageskip y))
    (`(p . (((class . "center ")) .
	    (,_ . ((b . (,_ . (,_ . ((span ((class . "smallcaps")) ,y) . ,_))))
		   . ,_))))
     (progn (string-match "^[0-9]*" y)
	    (list :b "sezione" (list (substring y 0 (match-end 0))))))
    (`(p . (((class . "hang ")) . ,y)) (list :b "nota" y))
    (`(h1 . (((class . "backtitle ")) . ,y)) (list :b "titolo" y))
    (_ (push (list :1n x) gi/roth/errors))))

(defun gi/roth/sxhtml-match-index (x)
  (pcase x
    ((pred null) (list :n))
    ((pred (lambda (z) (and (stringp z) (string-match-p "^[ \n\t]*$" z))))
     (list :n))
    (`(div . (((class . "timebreak ")) . ,_)) (list :n))
    (`(a . (((id . ,y)))) (list :a gi/roth/elt-pageskip y))
    (`(p . (((class . "index")) . ,y)) (list :b "voce indice" y))
    (`(p . (((class . "index-2")) . ,y)) (list :b "sottovoce indice" y))
    (`(p . (((class . "index-2") . ,y) . (,z . ,w)))
     (list :b "sottovoce indice"
	   (append w (cons (cons 'a y) (list z))) gi/roth/errors))
    (`(h1 . (((class . "backtitle ")) . ,y)) (list :b "titolo" y))
    (_ (push (list :1n x) gi/roth/errors))))

(defun gi/roth/sxhtml-elt (lst elt)
  (let ((lne-idx 0) (str-idx 0) (str ""))
    (dolist (x lst)
      (let* ((a (gi/roth/sxhtml-match-elt x))
	     (cat (nth 0 a)) (key (nth 1 a)) (txt (nth 2 a)))
	(pcase cat
	  (:n nil)
	  (:a (setq str-idx (+ str-idx (length txt))
		    str (concat str txt)))
	  (:b (let ((upd-idx (+ str-idx (length txt))))
		(puthash elt (cons (cons str-idx upd-idx) (gethash elt key))
			 key)
		(setq str-idx upd-idx
		      str (concat str txt))))
	  (:c (let ((key
		     (cond
		      ((string-match "^p[0-9]+$" txt) gi/roth/elt-pageskip)
		      ((string-match "^http" txt) gi/roth/elt-link)
		      ((string-match "px[0-9]+$" txt) gi/roth/elt-notesref)
		      (t gi/roth/elt-indexref))))
		(puthash elt (cons (cons str-idx txt) (gethash elt key))
			 key))))))
    (puthash elt (length str) gi/roth/elt)
    (let ((sentence-end-double-space nil) (m '(0)))
      (if (not (string-match (sentence-end) str))
	  (setq m (cons (length str) m))
	(while (string-match (sentence-end) (substring str (car m)))
	  (setq m (cons (+ (car m) (match-end 0)) m))))
      ;; non sentence-end terminating strings
      (when (not (= (car m) (length str))) (setq m (cons (length str) m)))
      ;; sentence-end exceptions
      (dolist (x (butlast (cdr m)))
	(let ((s (substring str 0 x)))
	  (dolist (y gi/roth/sentence-end-exceptions)
	    (when (and (>= (length s) (length y))
		       (string-suffix-p (concat y " ") s))
	      (setq m (remq x m))))))
      (dolist (x (reverse (cl-mapcar
			   (lambda (a b) (cons b a)) (butlast m) (cdr m))))
	(let ((lne (append elt (list lne-idx))))
	  (puthash lne x gi/roth/lne)
	  (puthash lne (substring str (car x) (cdr x)) gi/roth/lne-text)
	  (setq lne-idx (1+ lne-idx)))))))

(defun gi/roth/sxhtml-match-elt (x)
  (pcase x
    ((pred null) (list :n))
    ((pred (lambda (z) (and (stringp z) (string-match-p "^[ \n\t]*$" z))))
     (list :n))
    ((pred stringp) (list :a nil x))
    (`(span . (((class . "smallcaps")) ,y)) (list :a nil (downcase y)))
    (`(em ,_ (span ,_ ,y)) (list :b gi/roth/elt-italic (downcase y)))
    (`(em ,_ ,y) (list :b gi/roth/elt-italic y))
    (`(b ,_ ,y) (list :b gi/roth/elt-bold y))
    (`(a . (((id . ,y)))) (list :c nil y))
    (`(a . (((href . ,y)) . ,_)) (list :c nil y))
    (`(a . ((href . ,y))) (list :c nil y))
    (_ (push (list :2 x) gi/roth/errors))))

(defun gi/roth/elt-lines (elt)
  (let ((lne 0) (lnes ()))
    (while (gethash (append elt (list lne)) gi/roth/lne)
      (setq lnes (cons (append elt (list lne)) lnes) lne (1+ lne)))
    lnes))

;;;; exceptions

(defconst gi/roth/sentence-end-exceptions
  '("Ph.D." "Dr." "D.C." "a.m." "U.S."  " no." " No." " ed." " bk." "
    vol." " chap." "para."  " A." " B." " C." " D." " E." " F." " G."
    " H." " I." " J." " K." " L."  " M." " N." " O." " P." " Q." " R."
    " S." " T." " U." " V." " W." " X."  " Y." " Z."))

(defconst gi/roth/titles16
  '("The Kidney Donor’s Diet and Exercise Regime: You, Too, Can Qualify"))

(defconst gi/roth/titles18
  '("“Price Coherence and Adverse Intermediation”"
    "Charlie’s Place: The Saga of an American Frontier Homestead"
    "“Jumping the Gun.”"
    "“The High-Frequency Trading Arms Race: Frequent Batch Auctions as a Market Design Response”"
    "“Information Frictions and the Law of One Price: ‘When the States and the Kingdom Became United’”"
    "Judge John D. Bates to All United States Judges"
    "Two-Sided Matching: A Study in Game-Theoretic Modeling and Analysis"
    "“From Boston to Chinese Parallel to Deferred Acceptance: Theory and Experiments on a Family of School Choice Mechanisms”"
    "The Early Admissions Game"
    "“Propose with a Rose?"
    "Signaling in Internet Dating Markets"
    "The Handicap Principle: A Missing Piece of Darwin’s Puzzle"
    "The Histories"
    "Putting Auction Theory to Work"
    "“More Money, More Problems?"
    "Can High Pay Be Coercive and Repugnant?"
    "The Kidney Sellers: A Journey of Discovery in Iran"
    "When Altruism Isn’t Enough: The Case for Compensating Kidney Donors"
    "The Collected Works of F. A. Hayek"
    "The Road to Serfdom: Text and Documents—The Definitive Edition"
    "New Studies in Philosophy, Politics, Economics and the History of Ideas"))

(defconst gi/roth/common
  '("be" "a" "in" "it" "have" "as" "t" "s" "but" "are" "on" "can" "at"
    "one" "by" "make" "i" "or" "so" "who" "an" "more" "other" "some"
    "about" "he" "not" "many" "even" "also" "might" "get" "there" "all"
    "first" "just" "do" "only" "time" "each" "people" "hi" "like"
    "before" "new" "most" "way" "take" "may" "want" "will" "good" "such"
    "no" "high" "out" "use" "too" "after" "need" "early" "know" "any"
    "much" "up" "help" "place" "very" "kind" "don" "well" "another"
    "find" "come" "still" "try" "now" "hard" "making" "become" "go"
    "say" "safe" "then" "while" "give" "enough" "re" "allow" "better"
    "why" "d" "through" "being" "isn" "part" "often" "turn" "start"
    "big" "best" "long" "lot" "wait" "look" "over" "begin" "already"
    "think" "case" "less" "play" "fact" "us" "tell" "let" "later"
    "someone" "me" "own" "waiting" "consider" "happen" "between" "able"
    "little" "fail" "apply" "things" "instead" "true" "move" "possible"
    "once" "found" "earlier" "top" "bad" "back" "provide" "every" "old"
    "few" "almost" "getting" "down" "lots" "ways" "ad" "yet" "keep"
    "going" "notice" "longer" "least" "must" "last" "show" "set" "using"
    "put" "care" "easy" "thus" "run" "mean" "close" "looking" "soon"
    "though" "whole" "next" "given" "small" "clear" "again" "always"
    "under" "suppose" "suggest" "rather" "either" "near" "far" "around"
    "side" "together" "here" "off" "bring" "appear" "real" "thing"
    "never" "slow" "prove" "eventually" "higher" "quite" "agree" "away"
    "m" "seem" "really" "saw" "starting" "living" "p" "maybe" "late"
    "felt" "sent" "finding" "further" "main" "full" "strong" "feel"
    "taking" "certain" "thinking" "indeed" "parts" "pretty" "ahead"
    "ever" "sure" "lead" "led" "fall" "left" "ago" "ready" "various"
    "giving" "badly"))

;;; glossary

(defun gi/roth/glossary ()
  (gi/roth/gls :titles16 16) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :titles18 18) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :nprefixes 18 0) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :titles 18) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls-index) (gi/roth/gls-rev 'gi/roth/gls-index)
  (gi/roth/gls :specials) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :acronyms) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :numbers) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :names) (gi/roth/gls-rev 'gi/roth/gls-names)
  (gi/roth/gls :words) 
  (maphash (lambda (k v) (when (= (length v) 1) (remhash k gi/roth/gls)))
	   gi/roth/gls)
  (mapc (lambda (x) (remhash x gi/roth/gls)) gi/roth/common)
  (gi/roth/gls-rev 'gi/roth/gls))

(defun gi/roth/gls-index-regexps ()
  (let ((case-fold-search nil) (lst ()))
    (maphash
     (lambda (k v) (when (equal v "voce indice")
		(let* ((lne (append k (list 0)))
		       (x (gethash lne gi/roth/lne-text))
		       (y (if (string-match "^\\(?1:.+?\\)[., ]+$" x)
			      (match-string 1 x) x)))
		  (push (cons k (list y)) lst))))
     gi/roth/elt-type)
    (mapcar
     (lambda (x)
       (let ((a (cadr x)) (b ()))
	 (cond
	  ((string-match "“[^“”]+”" a)
	   (setq b (list (concat (substring (match-string 0 a) 0 -1)
				 "\\([.,:]\\)*”"))))
	  ((string-match "^\\(?1:[^“”]+\\), \\(?2:[^“”]+\\)(.+)$" a) 
	   (setq b (list (concat (match-string 2 a) (match-string 1 a)))))
	  ((string-match "^\\(?1:[^“”]+\\) (.+)$" a)
	   (setq b (list (match-string 1 a))))
	  ((string-match "^\\(?1:[^“”]+\\), \\(?2:[^“”]+\\)$" a)
	   (setq b (list (concat (match-string 2 a) " " (match-string 1 a))
			 (match-string 1 a))))
	  (t (setq b nil)))
	 (when (string-match "[[:lower:]]" (substring a 0 1))
	   (setq b (cons (concat (upcase (substring a 0 1))
				 (substring a 1 nil))
			 b)))
	 (cons (car x) (mapcar (lambda (y) (concat "\\(^\\|[^[:word:]]\\)\\(?1:" y
					      "\\)\\([^[:word:]]\\|$\\)"))
			       (cons a b)))))
     lst)))

(defconst gi/roth/gls-regexps
  `((:nprefixes . "\\(?1:^.+?: \\).+$")
    (:titles16  . ,(concat "\\(?1:" (mapconcat #'regexp-quote
					       gi/roth/titles16 "")
			   "\\)"))
    (:titles18  . ,(concat "\\(?1:" (mapconcat #'regexp-quote
					       gi/roth/titles18 "\\|")
			   "\\)"))
    (:titles    . "“\\(?1:[^“”]+\\),”")
    (:specials  . ,(concat "\\([[:blank:]]\\|^\\)\\(?1:\\("
			   "Ph\\.D\\.\\|Dr\\.\\|D\\.C\\.\\|a\\.m\\.\\|"
			   "U\\.S\\.\\|no\\.\\|No\\.\\|ed\\.\\|bk\\.\\|"
			   "vol\\.\\|chap\\.\\|para\\."
			   "\\)\\)\\([^[:word:]]\\|$\\)"))
    (:names     . ,(concat "[[:blank:]]"
			   "\\(?1:\\([[:upper:]]\\([.]\\|[[:lower:]]+\\)\\)+"
			   "\\([-[:blank:]]"
			   "[[:upper:]]\\([.]\\|[[:lower:]]+\\)\\)*\\)"
			   "\\([^[:word:]]\\|$\\)"))
    (:acronyms  . ,(concat "\\([[:blank:]]\\|^\\)\\(?1:\\(" ; and roman nums
			   "[[:upper:]]+[[:upper:]&]*[[:upper:]]+"
			   "\\)\\)\\([^[:word:]]\\|$\\)"))
    (:numbers   . ,(concat "\\([[:blank:]]\\|^\\)"
			   "\\(?1:[[:digit:]:,$]*[-]*[[:digit:]]+s?\\)"
			   "\\([^[:word:]]\\|$\\)"))
    (:words     . "\\(?1:[[:word:]-]+\\)")))

(defun gi/roth/gls-index ()
  (let ((case-fold-search nil))
    (maphash
     (lambda (k v)
       (let ((mtc ()) (flt ()) ; (((chk elt) strt end) ...) for line k
	     (b (gi/roth/protect k 'gi/roth/gls-names-r)))
	 (mapc
	  (lambda (x) (mapc
		  (lambda (y) (let ((a (gi/roth/match v (apply-partially
						    #'gi/roth/matcher y)
						 b)))
			   (mapc (lambda (z) (push (cons (car x) z) mtc)) a)))
		  (cdr x)))
	  (gi/roth/gls-index-regexps))
	 (dolist (y mtc)
	   (dolist (z (cdr (memq y mtc)))
	     (when (and (not (member y flt))
			(gi/roth/overlap-p
			 (nth 1 y) (nth 2 y) (nth 1 z) (nth 2 z)))
	       ;; tiebreak is quasi random
	       (if (<= (- (nth 2 y) (nth 1 y)) (- (nth 2 z) (nth 1 z)))
		   (push y flt) (push z flt)))))
	 (mapc
	  (lambda (x) (puthash (car x) (cons (list (append k (cdr x)) ())
					(gethash (car x) gi/roth/gls-index))
			  gi/roth/gls-index))
	  (cl-set-difference mtc flt :test #'equal))))
     gi/roth/lne-text)))

(defun gi/roth/gls (typ &optional chk lne)
  (maphash
   (lambda (k v)
     (when (and (not (member (gethash (butlast k) gi/roth/elt-type)
			     (pcase typ
			       ((or :words :numbers) ())
			       (_ '("titolo" "sezione" "capitolo" "parte")))))
		(if chk (= (car k) chk) t)
		(if lne (= (nth 2 k) lne) t))
       (let ((mtc (gi/roth/match
		   v (apply-partially #'gi/roth/matcher
				      (cdr (assoc typ gi/roth/gls-regexps)))
		   (apply (apply-partially #'gi/roth/protect k)
			  (pcase typ
			    (:nprefixes nil)
			    (:titles '(gi/roth/gls-names-r))
			    (_ '(gi/roth/gls-index-r gi/roth/gls-names-r)))))))
	 (mapc (lambda (x) (let ((key (pcase typ
				   (:words (gi/roth/root
					    (substring v (car x) (cadr x))))
				   (_ (substring v (car x) (cadr x)))))
			    (tbl (pcase typ
				   (:words gi/roth/gls)
				   (_ gi/roth/gls-names))))
			(when key (puthash key (cons (list (append k x) ())
						     (gethash key tbl))
					   tbl))))
	       (remq nil mtc)))))
   gi/roth/lne-text))

;; [str fun ((strt end) ...)] -> ((strt end) ...)
;; fun : str -> ((strt end) ...)
(defun gi/roth/match (str fun exc)
  (let ((mtc ()))
    (cl-mapc (lambda (x y)
	       (let ((a (funcall fun (substring str (cadr x) (car y)))))
		 (mapc (lambda (z) (when a (push (list (+ (cadr x) (car z))
						  (+ (cadr x) (cadr z)))
					    mtc)))
		       a)))
	     (cons '(nil 0) exc) (append exc `((,(length str)))))
    mtc))

(defun gi/roth/matcher (rgx str)
  (let ((case-fold-search nil) (adj 0) (mtc ()))
    (while (string-match rgx str)
      (push (list (+ adj (match-beginning 1)) (+ adj (match-end 1))) mtc)
      (setq str (substring str (match-end 0)) adj (+ adj (match-end 0))))
    mtc))

;; (lne (tbl ...)) -> ((strt end) ...)
(defun gi/roth/protect (lne &rest rtbls)
  (let ((nts ()))
    (mapc (lambda (x) (let ((tbl (symbol-value
			     (intern-soft (substring (symbol-name x) 0 -2)))))
		   (mapc (lambda (y) (mapc (lambda (z)
					(when (equal lne (butlast (car z) 2))
					  (push z nts)))
				      (gethash y tbl)))
			 (gethash lne (symbol-value x)))))
	  rtbls)
    (when (car nts) (sort (delete-dups (mapcar (lambda (x) (nthcdr 3 (car x))) nts))
			  (lambda (a b) (< (car a) (car b)))))))

(defun gi/roth/overlap-p (ba ea bb eb)
  "True if right-open integers interval BA-EA overlaps BB-EB"
  (or (and (>= ba bb) (< ba eb)) (and (> ea bb) (<= ea eb))
      (and (< ba bb) (> ea eb))))

(defun gi/roth/root (str)
  "Get wordnet root of STR"
  (let ((wn (with-temp-buffer
	      (apply 'call-process "wn" nil t nil `(,str "-over"))
	      (buffer-string))))
    (when (and wn (string> wn ""))
      (string-match "Overview.+ \\([[:word:]-]+\\)" wn)
      (match-string 1 wn))))

; ((key val)) -> (lne (key))
(defun gi/roth/gls-rev (tbl)
  (let ((rtbl (symbol-value (intern-soft (concat (symbol-name tbl) "-r")))))
    (clrhash rtbl)
    (maphash
     (lambda (key val)
       (mapc (lambda (x) (puthash (butlast (car x) 2)
			     (cons key (gethash (butlast (car x) 2) rtbl))
			     rtbl))
	     val))
     (symbol-value tbl))))

;;; unused

(defun gi/roth/key<= (a b)
  (cond ((< (nth 0 a) (nth 0 b)) t)
	((= (nth 0 a) (nth 0 b))
	 (cond ((< (nth 1 a) (nth 1 b)) t)
	       ((= (nth 1 a) (nth 1 b))
		(or (= (length a) 2)
		    (cond ((<= (nth 2 a) (nth 2 b)) t)
			  (t nil))))
	       (t nil)))
	(t nil)))

;;; complete setup

(defun gi/roth/setup ()
  (let* ((files (list gi/roth/roth gi/roth/glossary
		      gi/roth/rglossary gi/roth/traduzione))
	 (tables (list gi/roth/elt-ttables
		       gi/roth/elt-tables gi/roth/gls-tables
		       gi/roth/gls-rtables gi/roth/trd-tables))
	 (printout (cdr tables)))
    (setq gi/roth/errors ())
    (mapc (lambda (x) (unless (file-exists-p x) (write-region "" nil x))) files)
    (mapc (lambda (x) (mapc #'gi/roth/hash x)) tables)
    (mapc
     (lambda (x) (mapc (lambda (y) (clrhash (symbol-value (intern-soft (car y))))) x))
     tables)
    (gi/roth/xhtml) (gi/roth/fix-repeated-pageskips) (gi/roth/glossary) 
    (cl-mapc #'gi/roth/print printout files)))

(defun gi/roth/hash (alst)
  (intern (car alst))
  (set (intern-soft (car alst))
       (make-hash-table :test 'equal :size (cdr alst))))

(defun gi/roth/print (tbls fle)
  (let ((tblsyms (mapcar (lambda (x) (intern-soft (car x))) tbls)))
    (with-current-buffer (find-file-noselect fle)
      (erase-buffer)
      (cl-mapc (lambda (tab sym) (insert "(setq " (prin1-to-string sym)
				    " " (prin1-to-string tab) ")\n\n"))
	       (mapcar #'symbol-value tblsyms) tblsyms)
      (save-buffer))))

(defun gi/roth/printt (ntbls tbls fle)
  (with-current-buffer (find-file-noselect fle)
    (erase-buffer)
    (cl-mapc (lambda (tab ssym) (insert "(setq " ssym
				  " " (prin1-to-string tab) ")\n\n"))
	     ntbls (mapcar #'car tbls))
    (save-buffer)))

(defun gi/roth/load ()
  (let ((files (list gi/roth/roth gi/roth/glossary
		     gi/roth/rglossary gi/roth/traduzione)))
    (mapc #'load-file files)))

;;; ugly

(defun gi/roth/fix-repeated-pageskips ()
  (let ((rem ()))
    (maphash
     (lambda (k v) (let ((a (gethash (gi/roth/following k) gi/roth/lne-pageskip)))
		(when (equal (cdar v) (cdar a))
		  (push (gi/roth/following k) rem))))
     gi/roth/lne-pageskip)
    (mapc (lambda (x) (remhash x gi/roth/lne-pageskip)) rem)))

;;; order

(defun gi/roth/following (lne)
  (let ((nlne (list (nth 0 lne) (nth 1 lne) (1+ (nth 2 lne))))
	(nelt (list (nth 0 lne) (1+ (nth 1 lne)) 0))
	(nchk (list (1+ (nth 0 lne)) 0 0)))
    (or (and (gethash nlne gi/roth/lne-text) nlne)
	(and (gethash nelt gi/roth/lne-text) nelt)
	(and (gethash nchk gi/roth/lne-text) nchk))))

(defun gi/roth/previous (lne)
  (cond
   ((> (nth 2 lne) 0) (append (butlast lne) (list (- (nth 2 lne) 1))))
   ((> (nth 1 lne) 0)
    (let ((x (list (nth 0 lne) (- (nth 1 lne) 1) 0)))
      (while (not (equal (gi/roth/following x) lne))
	(setq x (gi/roth/following x)))
      x))
   ((> (nth 0 lne) 0)
    (let ((x (list (- (nth 0 lne) 1) 0 0)))
      (while (not (equal (gi/roth/following x) lne))
	(setq x (gi/roth/following x)))
      x))
   (t '(0 0 0))))

;;; glossario
;; by construction, glossary match intervals do not overlap

(defvar gi/roth/g) ; (num ((strt end) tstr altstr ...) ...)
(defvar gi/roth/w) ; num

;; -> (((strt end) strt end) ...) | nil
(defun gi/roth/g-init (lne)
  (let* ((a (apply #'append
		  (mapcar (lambda (x) (gi/roth/g-init-2 lne (car x)))
			  gi/roth/gls-rtables)))
	 (trd (mapcar (lambda (x)
			(let ((b (car x)))
			  (when (cadr b) (cons (nthcdr 3 (car b)) (cdr b)))))
		      a)))
    (setq gi/roth/g (cons 0 (sort
			     (copy-sequence
			      (mapcar (lambda (x) (gi/roth/g-init-1 x)) a))
			     (lambda (a b) (< (caar a) (caar b)))))
	  gi/roth/w (length (gethash lne gi/roth/gls-r)))
    (if (and (car trd) (not (= (length (cdr gi/roth/g)) (length trd))))
	(error "incomplete translation") trd)))

;; -> (lst ...)
(defun gi/roth/g-init-1 (lst)
  (cons (nthcdr 3 (caar lst))
	(cons (or (gi/roth/g-init-trad (car lst))
		  (gi/roth/g-init-word (car lst)))
	      (remq nil (mapcar (lambda (x) (gi/roth/g-init-trad x))
				(cdr lst))))))

;; -> ((((chk elt lne strt end) strt end) ...) ...)
(defun gi/roth/g-init-2 (lne tblstr)
  (let* ((rtbl (symbol-value (intern-soft tblstr)))
	 (tbl (symbol-value (intern-soft (substring tblstr 0 -2))))
	 (mtc (mapcar (lambda (x) (gethash x tbl)) (gethash lne rtbl)))
	 (otc (mapcar (lambda (x) (mapcar
			      (lambda (y) (when (equal (butlast (car y) 2) lne)
				       (cons y (remq y x))))
			      x))
		      mtc)))
    (apply #'append (cl-remove-duplicates 	; multiple occurrencies
		     (mapcar (lambda (x) (remq nil x)) otc)
		     :test #'equal))))

;; -> ((chk elt lne strt end) strt end) -> str
(defun gi/roth/g-init-word (ent)
  (substring (gethash (butlast (car ent) 2) gi/roth/lne-text)
	     (nth 3 (car ent)) (nth 4 (car ent))))

;; -> ((chk elt lne strt end) strt end) -> str
(defun gi/roth/g-init-trad (ent)
  (when (cadr ent)
    (substring (gethash (butlast (car ent) 2) gi/roth/trd-text)
	       (cadr ent) (car (cddr ent)))))

(defun gi/roth/g (kyw)
  (let* ((a (nth (car gi/roth/g) (cdr gi/roth/g))))
    (pcase kyw
      (:current (cadr a))
      (:current-id (car a))
      (:raw a)
      (:list (cdr a))
      (:word? (< (car gi/roth/g) gi/roth/w)))))

;;; edizione

(defvar gi/roth/current-line)
(defvar gi/roth/buffers-windows)

;; faces:
;; highlight	 -> current glossary
;; bold		 -> non current glossary
;; warning	 -> unmatched refs
;; success	 -> matched refs
;; add underline -> italic
;; match         -> temp overlays
(defconst gi/roth/color-borders "black")

(defun gi/roth/split ()
  (let* ((names '("edit" "curr" "xtra" "trsl" "glos" "prev" "succ"))
	 (w1 (selected-window)) (w2 (split-window w1 -78 'left))
	 (w3 (split-window w1 -78 'right)) (w4 (split-window w1 -24 'above))
	 (w5 (split-window w1 -24 'below)) (w6 (split-window w2 -24 'above))
	 (w7 (split-window w2 -24 'below)))
    (setq cursor-in-non-selected-windows nil
	  mode-line-format nil
	  ;max-mini-window-height 0
	  )
    (set-frame-parameter nil 'right-divider-width 2)
    (set-frame-parameter nil 'bottom-divider-width 2)
    (set-face-attribute 'window-divider nil :foreground gi/roth/color-borders)
    (setq gi/roth/buffers-windows
	  (cl-mapcar (lambda (x y) (set-window-buffer x y)
		       (with-current-buffer y
			 (setq cursor-in-non-selected-windows nil
			       mode-line-format nil
			       max-mini-window-height 0))
		       (cons x y))
		     (list w1 w2 w3 w4 w5 w6 w7)
		     (mapcar
		      (lambda (x) (let* ((bsym (intern
					   (concat "gi/roth/" x "-buf")))
				    (bnme (concat "roth-" x)))
			       (set bsym (get-buffer-create bnme))
			       (symbol-value bsym)))
		      names)))))

(defun gi/roth/buffers (lne)
  (setq gi/roth/current-line lne
	gi/roth/edit-italic ()
	gi/roth/edit-notesref ()
	gi/roth/edit-indexref ()
	gi/roth/edit-link ())
  (let ((inhibit-read-only t)
	(trd (gi/roth/g-init lne))
	(str (gethash lne gi/roth/trd-text)))
    (gi/roth/line lne)
    (gi/roth/xtra-toggle)
    (with-selected-window (car (rassoc gi/roth/trsl-buf
				       gi/roth/buffers-windows))
      (with-current-buffer gi/roth/trsl-buf
	(erase-buffer)
	(let ((l (gi/roth/previous lne)) (ll nil))
	  (dotimes (x 10)
	    (goto-char (point-min))
	    (if (not (equal l ll)) (insert (concat "\n\n" (gi/roth/out l)))
	      (insert "\n\n"))
	    (setq ll l l (gi/roth/previous l)))
	  (fill-region (point-min) (point-max))
	  (goto-char (- (point-max) 1)) (recenter -1))))
    (gi/roth/word)
    (with-selected-window (car (rassoc gi/roth/prev-buf
				       gi/roth/buffers-windows))
      (with-current-buffer gi/roth/prev-buf
	(erase-buffer)
	(let ((l (gi/roth/previous lne)))
	  (dotimes (x 10)
	    (goto-char (point-min))
	    (when (not (equal l lne))
	      (insert (concat "\n[" (number-to-string (nth 0 l)) " "
			      " " (number-to-string (nth 1 l)) " "
			      " " (number-to-string (nth 2 l)) "] "
			      (gethash l gi/roth/lne-text) "\n"))
	      (setq l (gi/roth/previous l))))
	  (fill-region (point-min) (point-max))
	  (goto-char (- (point-max) 1)) (recenter -1)
	  (font-lock-add-keywords nil '(("\\[.*\\]" . 'bold)))
	  (font-lock-fontify-buffer))))
    (with-selected-window (car (rassoc gi/roth/succ-buf
				       gi/roth/buffers-windows))
      (with-current-buffer gi/roth/succ-buf
	(erase-buffer)
	(let ((l (gi/roth/following lne)))
	  (dotimes (x 10)
	    (goto-char (point-max))
	    (when l (insert (concat "\n[" (number-to-string (nth 0 l)) " "
				    " " (number-to-string (nth 1 l)) " "
				    " " (number-to-string (nth 2 l)) "] "
				    (gethash l gi/roth/lne-text) "\n"))
		  (setq l (gi/roth/following l))))
	  (fill-region (point-min) (point-max))
	  (goto-char (point-min)) (recenter 1)
	  (font-lock-add-keywords nil '(("\\[.*\\]" . 'bold)))
	  (font-lock-fontify-buffer))))
    (with-selected-window (car (rassoc gi/roth/edit-buf
				       gi/roth/buffers-windows))
      (with-current-buffer gi/roth/edit-buf
	(setq header-line-format "")
	(gi/roth/keys) (use-local-map gi/roth/keys)
	(erase-buffer)
	(if str (let* ((ita (gi/roth/str-buf
			     (gethash lne gi/roth/trd-italic) :buf))
		       (nt0 (gi/roth/str-buf
			     (gethash lne gi/roth/lne-notesref) :buf))
		       (nte (when nt0 (mapcar (lambda (x) (cons :note x)) nt0)))
		       (id0 (gi/roth/str-buf
			     (gethash lne gi/roth/lne-indexref) :buf))
		       (idx (when id0 (mapcar (lambda (x) (cons :indx x)) id0)))
		       (ln0 (gi/roth/str-buf
			     (gethash lne gi/roth/lne-link) :buf))
		       (lnk (when ln0 (mapcar (lambda (x) (cons :link x)) ln0)))
		       (ref (append nte idx lnk)))
		  (dolist (x trd)
		    (setq str (gi/roth/propertize
			       str (cadr x) (car (cddr x)) (car x) :glossary)))
		  (insert str)
	      ;; insert markers
		  (when (car ita) (dolist (x ita) (gi/roth/edit-italic x)))
		  (when (car ref) (dolist (y ref) (gi/roth/edit-refs
						   (car y) (cdr y)))))
	  (mapc (lambda (x) (insert (concat " " (gi/roth/propertize
					    (cadr x) 0 (length (cadr x)) (car x)
					    :glossary))))
		(sort (cdr (copy-sequence gi/roth/g))
		      (lambda (a b) (< (caar a) (caar b))))))))))

(defun gi/roth/line (lne)
  (let ((inhibit-read-only t)
	(str (gethash lne gi/roth/lne-text))
	(lst (apply #'append
		 (mapcar (lambda (y) (gethash lne y))
			 (list gi/roth/lne-pageskip gi/roth/lne-indexref
			       gi/roth/lne-notesref gi/roth/lne-link))))
	(adj 0)
	(ovs ()))
    (dolist (x (cdr gi/roth/g))
      (setq str (gi/roth/propertize str (caar x) (car (cdar x)) (car x)
				    :glossary)))
    (dolist (x (gethash lne gi/roth/lne-italic))
      (put-text-property (car x) (cdr x) :corsivo t str))
    (dolist (x lst)
      (setq str (concat (substring str 0 (+ adj (car x)))
			(gi/roth/propertize (concat "[" (cdr x) "]")
					    0 (+ 2 (length (cdr x)))
					    (car x) :ref)
			(substring str (+ adj (car x)) nil))
	    adj (+ 2 (length (cdr x)) adj)))
    (with-selected-window (car (rassoc gi/roth/curr-buf
				       gi/roth/buffers-windows))
      (with-current-buffer gi/roth/curr-buf
	(setq header-line-format "")
	(erase-buffer)
	(insert str)
	(mapc (lambda (x) (overlay-put x 'face 'match))
	      (gi/roth/get-text-property-corsivo))
	(fill-region (point-min) (point-max))))))

(defun gi/roth/line-page (lne)
  (or (cdr (gi/roth/line-pageskip lne))
      (let ((x (gi/roth/previous lne)))
	(while (not (gi/roth/line-pageskip x))
	  (setq x (gi/roth/previous x)))
	(cdr (gi/roth/line-pageskip x)))))

;; no line contains more than two pageskips, and when a line contains
;; two pageskips, one refers to a blank page
(defun gi/roth/line-pageskip (lne)
  (let ((skps (gethash lne gi/roth/lne-pageskip)))
    (or (and skps (> (length skps) 1)
	     (string> (cddr skps) (cdar skps)) (cdr skps))
	(car skps))))

(defun gi/roth/info (lne)
  (let* ((i (gi/roth/following lne)) (nelt) (nlne)
	 (pge (gi/roth/line-page lne))
	 (frq (gi/roth/words-occurrencies lne))
	 (ind (gi/roth/page-index pge)))
    (while i (cond ((and (not nlne) (not (= (nth 1 i) (nth 1 lne))))
		    (setq nlne (max (nth 2 (gi/roth/previous i)) (nth 2 lne))
			  i (gi/roth/following i)))
		   ((and nlne (not nelt) (not (= (nth 0 i) (nth 0 lne))))
		    (setq nelt (max (nth 1 (gi/roth/previous i)) (nth 1 lne))
			  i (gi/roth/following i)))
		   ((and nlne nelt) (setq i nil))
		   ((not (gi/roth/following i))
		    (setq nelt (nth 1 i) nlne (nth 2 i)))
		   (t (setq i (gi/roth/following i)))))
    (when (equal lne '(20 1 1)) (setq nlne 1 nelt 1))
    (concat "\n"
	    (format "sezione  : %d (20)\n" (nth 0 lne))
	    (format "elemento : %d (%d)\n" (nth 1 lne) nelt)
	    (format "linea    : %s (%d)\n" (nth 2 lne) nlne)
	    (format "pagina   : %s\n" pge)
	    (format "tipo     : %s\n" (gethash (butlast lne) gi/roth/elt-type))
	    "\nfrequenze\n\n"
	    (if frq (mapconcat (lambda (x) (format "%18s : %d" (car x) (cdr x)))
			       frq "\n") "...")
	    "\n\nindice\n\n"
	    (if ind
		(mapconcat
		 (lambda (a) (concat "-> " (nth 0 a)
				(when (nth 1 a) (concat "[" (nth 1 a) "]"))
				(when (nth 2 a) (concat " <- " (nth 2 a)))
				(when (nth 3 a) (concat "[" (nth 3 a) "]"))))
		 (mapcar
		  (lambda (x)
		    (mapcar
		     (lambda (y) (and (stringp y)
				 (replace-regexp-in-string "\\(, \\)*$" "" y)))
		     x))
		  ind)
		 "\n\n") "..."))))

(defun gi/roth/words-occurrencies (lne)
  (mapcar (lambda (x) (cons x (length (gethash x gi/roth/gls))))
	  (gethash lne gi/roth/gls-r)))

(defun gi/roth/page-index (ref)
  (let ((ind ()))
    (maphash
     (lambda (k v)
       (when (cl-some (lambda (x) (string-match (concat "#" ref "$") (cdr x))) v)
	 (push k ind)))
     gi/roth/lne-indexref)
    (mapcar
     (lambda (x)
       (append
	(list (gethash x gi/roth/lne-text) (gethash x gi/roth/trd-text))
	(if (equal (gethash (butlast x) gi/roth/elt-type)
		   "sottovoce indice")
	    (let ((i (gi/roth/previous x)))
	      (while (and i (equal (gethash (butlast i) gi/roth/elt-type)
				   "sottovoce indice"))
		(setq i (gi/roth/previous i)))
	      (list (gethash i gi/roth/lne-text) (gethash i gi/roth/trd-text)))
	  ())))
     ind)))

(defun gi/roth/word ()
  (with-selected-window (car (rassoc gi/roth/glos-buf
				     gi/roth/buffers-windows))
    (with-current-buffer gi/roth/glos-buf
      (erase-buffer) (insert "\n")
      (let* ((inhibit-read-only t)
	     (str "")
	     (lst (delete-dups (gi/roth/g :list)))
	     (top (nth 0 lst))
	     (len (length lst))
	     (lim (> len 20))
	     (ita (format "[corsivo: %d/%d]"
			  (length gi/roth/edit-italic)
			  (length (gethash gi/roth/current-line
					   gi/roth/lne-italic))))
	     (cid (gi/roth/g :current-id))
	     (roo (if (gi/roth/g :word?) 
		      (format "[root: %s]"
			      (gi/roth/root
			       (substring
				(gethash gi/roth/current-line
					 gi/roth/lne-text)
				(car cid) (cadr cid)))) "")))
	(dotimes (x (if lim 20 (max 2 len)))
	  (setq str
		(cond
		 ((= x 0)
		  (concat (format
			   "%39s" (gi/roth/propertize top 0 (length top)
						      nil :glossary))
			  (format "%39s\n" ita)))
		 ((= x 1)
		  (concat (format "%39s" (or (nth 1 lst) ""))
			  (format "%39s\n" roo)))
		 (t (format "%39s\n" (or (nth x lst) "")))))
	  (setq str (concat str (if lim (format "%39s\n" "...") "")))
	  (insert str))))))

(defvar gi/roth/xtra-toggle 0)
(defun gi/roth/xtra-toggle ()
  (let ((gi/roth/xtra-toggle (if (not (boundp 'gi/roth/xtra-toggle)) 0
			       gi/roth/xtra-toggle)))
    (interactive)
    (with-selected-window (car (rassoc gi/roth/xtra-buf
				       gi/roth/buffers-windows))
      (with-current-buffer gi/roth/xtra-buf
	(erase-buffer)
	(pcase gi/roth/xtra-toggle
	  (0 (progn
		 (insert (gi/roth/info gi/roth/current-line))
		 (goto-char (point-min))
		 (fill-region (re-search-forward "^indice") (point-max))
		 (goto-char (point-min)) (recenter 1)
		 (font-lock-add-keywords nil '(("^indice" . 'bold)))
		 (font-lock-add-keywords nil '(("^frequenze" . 'bold)))
		 (font-lock-fontify-buffer)
		 (setq gi/roth/xtra-toggle (1+ gi/roth/xtra-toggle))))
	  (1 (progn
		 (if (gi/roth/g :word?)
		     (with-temp-buffer
		       (apply 'call-process "wn" nil t nil
			      `(,(gi/roth/g :current) "-over"))
		       (buffer-string))
		   (message "no wordnet root in glossary"))
		 (fill-region (point-min) (point-max))
		 (setq gi/roth/xtra-toggle (1+ gi/roth/xtra-toggle))))
	  (2 (let ((str "")
		     (occ (mapcar
			   (lambda (k v) (cons k (list (gethash (butlast (car v) 2)
							   gi/roth/lne-text)
						  (gethash (butlast (car v) 2)
							   gi/roth/trd-text))))
			   (gethash (gi/roth/g :current) gi/roth/gls))))
		 (dotimes (x (length occ))
		   (setq str (concat str (car x) " -->[ "
				     (or (cdr x) "...") " ]\n\n")))
		 (fill-region (point-min) (point-max))
		 (setq gi/roth/xtra-toggle (1+ gi/roth/xtra-toggle)))))))))

;; types <- :glossary :ref
(defun gi/roth/propertize (str beg end ref typ)
  (let ((str (copy-sequence str)))
    (if (equal typ :glossary)
	(put-text-property beg end 'face 'bold str)
      (put-text-property beg end 'face 'warning str))
    (put-text-property beg end :roth ref str)
    (put-text-property beg end 'front-sticky t str)
    (put-text-property beg end 'rear-nonsticky t str)
    (put-text-property beg end 'read-only t str)
    str))

(defun gi/roth/edit-glossary (&optional back)
  (interactive)
  (let ((inhibit-read-only t)
	(f (gi/roth/g :current-id))
	(l (length (cdr gi/roth/g)))
	(n (car gi/roth/g)))
    (if back (setcar gi/roth/g (if (= n 0) (- l 1) (- n 1)))
      (setcar gi/roth/g (if (= n (- l 1)) 0 (1+ n))))
    (gi/roth/update-bufs f (gi/roth/g :current-id))
    (gi/roth/word)))

(defun gi/roth/choose-word (&optional up)
  (interactive)
  (let* ((inhibit-read-only t)
	 (s (gi/roth/g :list))
	 (new (if up (cons (car (last s)) (butlast s))
		(append (cdr s) (list (car s))))))
    (setcdr (nth (car gi/roth/g) (cdr gi/roth/g)) new)
    (gi/roth/word)))

(defun gi/roth/edit-word (&optional new)
  (interactive)
  (when new (setcdr (nth (car gi/roth/g) (cdr gi/roth/g))
		    (cons (completing-read "-> " (gi/roth/g :list))
			  (gi/roth/g :list))))
  (let ((inhibit-read-only t)
	(str (gi/roth/g :current))
	(x 1))
    (with-current-buffer gi/roth/edit-buf
      (while (not (equal (gi/roth/g :current-id)
			 (get-text-property x :roth)))
	(setq x (1+ x)))
      (goto-char x)
      (delete-region x (next-single-property-change x :roth nil (point-max)))
      (insert (gi/roth/propertize str 0 (length str)
				  (gi/roth/g :current-id) :glossary))
      (gi/roth/update-bufs (gi/roth/g :current-id) (gi/roth/g :current-id))
      (gi/roth/word))))

(defun gi/roth/rm-current-word ()
  (interactive)
  (let* ((a (append gi/roth/current-line (gi/roth/g :current-id)))
	 (b (catch 'rm
	      (dolist (x (mapcar #'car gi/roth/gls-rtables))
		(dolist (y (gethash gi/roth/current-line (symbol-value
							  (intern-soft x))))
		  (dolist (z (gethash y (symbol-value (intern-soft
						       (substring x 0 -2)))))
		    (when (equal (car z) a)
		      (throw 'rm (cons (or (and (stringp y) y)
					   (prin1-to-string y))
				       x)))))))))
    (when (y-or-n-p (format "elimina %s da %s"
			    (car b) (substring (cdr b) 0 -2)))
      (remhash (car b) (symbol-value (intern-soft (substring (cdr b) 0 -2))))
      (gi/roth/gls-rev (intern-soft (substring (cdr b) 0 -2)))
      (cl-mapc #'gi/roth/print (list gi/roth/gls-tables gi/roth/gls-rtables)
	       (list gi/roth/glossary gi/roth/rglossary))
      (message nil) (gi/roth/buffers gi/roth/current-line))))

(defun gi/roth/guess-words (&optional revert) ; fixme
  (interactive)
  (let* ((a    ; which glossary items are words?
	  ())
	 (b    ; are there any already translated literal matches?
	  ())
	 (c    ; are the translations different?
	  ())
	 (d    ; map the most frequent translation, if any, to word items
	  ()))
    ()))
	  
(defun gi/roth/update-bufs (a b)
  (let ((inhibit-read-only t))
    (mapc (lambda (x)
	    (with-current-buffer x
	      (let ((from (gi/roth/get-text-property a))
		    (to (gi/roth/get-text-property b)))
		(put-text-property (car from) (cadr from) 'face 'bold)
		(put-text-property (car to) (cadr to) 'face 'highlight))))
	  (list gi/roth/curr-buf gi/roth/edit-buf))))

(defun gi/roth/update-buf-refs (pnt)
  (with-current-buffer gi/roth/curr-buf
    (let ((inhibit-read-only t)
	  (a (gi/roth/get-text-property pnt)))
      (put-text-property (car a) (cadr a) 'face 'success))))

(defun gi/roth/get-text-property (val &optional str)
  (let ((a (if str 0 1))
	(b (if str 0 1)))
    (while (not (equal (get-text-property a :roth (when str str)) val))
      (setq a (1+ a)))
    (setq b (next-single-property-change
	     a :roth (when str str) (if str (length str) (point-max))))
    (list a b)))

(defun gi/roth/get-text-property-corsivo (&optional str)
  (let ((a 1) (b 1) (ovs ()))
    (goto-char (point-min))
    (when (get-text-property a :corsivo)
      (setq b (next-single-property-change a :corsivo nil (point-max)))
      (push (make-overlay a b) ovs) (setq a b))
    (while (< a (point-max))
      (while (and (< a (point-max)) (not (get-text-property a :corsivo)))
	(setq a (1+ a)))
      (setq b (next-single-property-change a :corsivo nil (point-max)))
      (unless (= a b) (push (make-overlay a b) ovs))
      (setq a b))
    ovs))

(defun gi/roth/str-buf (lst sym)
  (let ((adj (if (equal sym :buf) 1 -1)))
    (mapcar
     (lambda (x) (cons (+ (car x) adj)
		  (or (and (numberp (cdr x)) (+ (cdr x) adj)) (cdr x))))
     lst)))

;;;; markers

(defvar gi/roth/edit-italic ())
(defun gi/roth/edit-italic (a &optional b)
  (interactive "r")
  (if (>= (length gi/roth/edit-italic)
	  (length (gethash gi/roth/current-line gi/roth/lne-italic)))
      (message "fully italicized")
    (let* ((inhibit-read-only t)
	   (beg (if b a (car a))) (end (if b b (cdr a)))
	   (begm (make-marker)) (endm (make-marker))
	   (ovr (make-overlay beg end)))
      (set-marker begm beg) (set-marker endm end)
      (push (cons beg end) gi/roth/edit-italic)
      (overlay-put ovr 'face 'match))))

(defvar gi/roth/edit-notesref ())
(defvar gi/roth/edit-link ())
(defvar gi/roth/edit-indexref ())
(defun gi/roth/edit-refs (kyw &optional obj)
  (interactive)
  (let* ((m (make-marker))
	 (a (cond ((equal kyw :note) gi/roth/lne-notesref)
		  ((equal kyw :link) gi/roth/lne-link)
		  (t gi/roth/lne-indexref)))
	 (b (cond ((equal kyw :note) (symbol-name 'gi/roth/edit-notesref))
		  ((equal kyw :link) (symbol-name 'gi/roth/edit-link))
		  (t (symbol-name 'gi/roth/edit-indexref))))
	 (c (if obj (cdr obj)
	      (completing-read
	       "-> " (remq nil (mapcar (lambda (x)
					 (unless
					     (cl-some
					      (lambda (y) (equal (cdr x) (cdr y)))
					      (symbol-value (intern-soft b)))
					   (cdr x)))
				       (gethash gi/roth/current-line a)))
	       nil t))))
    (set-marker m (if obj (car obj) (point)))
    (set (intern-soft b) (cons (cons m c) (symbol-value (intern-soft b))))
    (gi/roth/update-buf-refs
     (car (rassoc c (gethash gi/roth/current-line a))))))

(defun gi/roth/rm-markers ()
  (interactive)
  (mapc (lambda (x) (set-marker (car x) nil))
	(append gi/roth/edit-notesref gi/roth/edit-indexref gi/roth/edit-link))
  (mapc (lambda (x) (set x nil))
	'(gi/roth/edit-notesref gi/roth/edit-indexref gi/roth/edit-link)))

;;;; ndt

(defun gi/roth/ndt ()
  (interactive)
  (let ((a (read-from-minibuffer "-> ")))
    (puthash gi/roth/current-line
	     (cons a (gethash gi/roth/current-line gi/trd-note))
	     gi/roth/trd-note)))

;;; produzione

(defun gi/roth/out (lne)
  (let ((trd (gethash lne gi/roth/trd-text)))
    (if (not trd) (format " ...[%d %d %d]... "
			  (nth 0 lne) (nth 1 lne) (nth 2 lne))
      (gi/roth/format-frase lne))))

(defconst gi/roth/format 
  '((gi/roth/elt-type . (lambda (x) (format "[%s]" (upcase x))))
    (gi/roth/lne-pageskip . (lambda (x) (format "[PAGINA: %s]" x)))
    (gi/roth/trd-italic . (lambda () (cons "[CORSIVO]" "[FINE CORSIVO]")))
    (gi/roth/trd-notesref . (lambda (x) (format "[NOTA: %s]" x)))
    (gi/roth/trd-indexref . (lambda (x) (format "[RIFERIMENTO: %s]" x)))
    (gi/roth/trd-note . (lambda (x) (format "[NDT: %s]" x)))
    (gi/roth/trd-link . (lambda (x) (format "[URL: %s]" x)))))

(defun gi/roth/format-frase (lne)
  (let* ((str (gethash lne gi/roth/trd-text))
	 (elts ()) (len (length str)) (adj 0))
    (mapc (lambda (x)
	    (let* ((tblsym (car x))
		   (tbl (symbol-value tblsym))
		   (fun (cdr x))
		   (lst (or (and (= (nth 2 lne) 0)
				 (equal tblsym 'gi/roth/elt-type)
				 (list (gethash (butlast lne) tbl)))
			    (gethash lne tbl))))
	      (mapc 				; -> (precedence pnt str)
	       (lambda (y) (pcase tblsym
			(`gi/roth/elt-type
			 (push (list 1.0 0 (funcall fun y)) elts))
			(`gi/roth/lne-pageskip
			 (let ((arg (progn
				      (string-match
				       "p\\(?1:[[:digit:]iv]+$\\)" (cdr y))
				      (match-string 1 (cdr y)))))
			   (push (list
				  (or
				   (and (= (car y) 0)
					(/ (string-to-number arg) 1000.0))
				   9.0)
				  (or (and (= (car y) 0) 0)
				      (length str))
				  (funcall fun arg))
				 elts)))
			(`gi/roth/trd-italic
			 (let ((lab (funcall fun)))
			   (push (list 8.0 (car y) (car lab)) elts)
			   (push (list 2.0 (cdr y) (cdr lab)) elts)))
			(`gi/roth/trd-notesref
			 (let ((arg (progn
				      (string-match
				       "px\\(?1:[[:digit:]]+$\\)" (cdr y))
				      (match-string 1 (cdr y)))))
			   (push (list 3.0 (car y) (funcall fun arg)) elts)))
			(`gi/roth/trd-indexref
			 (let ((arg (gi/roth/fixed-indexref lne (cdr y))
				;; (progn
				;;       (string-match
				;;        "#p\\(?1:[[:digit:]]+$\\)" (cdr y))
				;;       (match-string 1 (cdr y)))
				))
			   (push (list 4.0 (car y) (funcall fun arg)) elts)))
			(`gi/roth/trd-note
			 (push (list 6.0 (car y) (funcall fun (cdr y))) elts))
			(`gi/roth/trd-link
			 (push (list 5.0 (car y) (funcall fun (cdr y))) elts))))
	       lst)))
	  gi/roth/format)
    (let ((ord (mapcar #'cdr (sort elts (lambda (a b)
					  (cond ((< (nth 1 a) (nth 1 b)) t)
						((< (nth 0 a) (nth 0 b)) t)
						(t nil)))))))
      (dolist (x ord)
	(setq str (concat (substring str 0 (+ adj (car x))) (cadr x)
			  (substring str (+ adj (car x)) nil))
	      adj (- (length str) len))))
    str))

;;; progresso

(defun gi/roth/line-trd (lne)
  (with-current-buffer gi/roth/edit-buf
    (let ((inhibit-read-only t)
	  (str (buffer-string))
	  (lnestr (mapconcat #'number-to-string lne " "))
	  (ita (gi/roth/str-buf gi/roth/edit-italic :str))
	  (nte (mapcar (lambda (x) (cons (- (marker-position (car x)) 1) (cdr x)))
		       gi/roth/edit-notesref))
	  (idx (mapcar (lambda (x) (cons (- (marker-position (car x)) 1) (cdr x)))
		       gi/roth/edit-indexref))
	  (lnk (mapcar (lambda (x) (cons (- (marker-position (car x)) 1) (cdr x)))
		       gi/roth/edit-link)))
      (if (and (and (= (length ita) (length (gethash lne gi/roth/lne-italic)))
		    (message "controllo linea [%s]: corsivo" lnestr))
	       (and (= (length nte) (length (gethash lne gi/roth/lne-notesref)))
		    (cl-every (lambda (x) (rassoc (cdr x) nte))
			      (gethash lne gi/roth/lne-notesref))
		    (message "controllo linea [%s]: note" lnestr))
	       (and (= (length idx) (length (gethash lne gi/roth/lne-indexref)))
		    (cl-every (lambda (x) (rassoc (cdr x) idx))
			      (gethash lne gi/roth/lne-indexref))
		    (message "controllo linea [%s]: indice" lnestr))
	       (and (= (length lnk) (length (gethash lne gi/roth/lne-link)))
		    (cl-every (lambda (x) (rassoc (cdr x) lnk))
			      (gethash lne gi/roth/lne-link))
		    (message "controllo linea [%s]: url" lnestr))
	       (and (mapc (lambda (x) (progn
				   (remhash lne (cdr x))
				   (mapc (lambda (y)
					   (puthash
					    lne (cons y (gethash lne (cdr x)))
					    (cdr x)))
					 (car x))))
			  (list (cons ita gi/roth/trd-italic)
				(cons nte gi/roth/trd-notesref)
				(cons idx gi/roth/trd-indexref)
				(cons lnk gi/roth/trd-link)))
		    (message "linea [%s]: elementi" lnestr))
	       (and (gi/roth/glossary-trd lne str)
		    (message "linea [%s]: glossario" lnestr))
	       (and (puthash lne (substring-no-properties str)
			     gi/roth/trd-text)
		    (message "linea [%s]: testo" lnestr)))
	  (message "linea [%s] completa" lnestr)
	nil))))

(defun gi/roth/glossary-trd (lne str)
  (catch 'err
    (mapc
     (lambda (x)
       (let ((rtbl (gethash lne (symbol-value (intern-soft x))))
	     (tbl (symbol-value (intern-soft (substring x 0 -2)))))
	 (dolist (key rtbl)
	   (let ((y (gethash key tbl)) (strt 0) (end 0))
	     (dolist (val y)
	       (when (equal (butlast (car val) 2) lne)
		 (let ((old (cdr val))
		       (oldtxt (gethash key gi/roth/trd-text))
		       (new (gi/roth/get-text-property
			     (nthcdr 3 (car val)) str)))
		   (or (and new (puthash key
					 (cons (cons (car val) new)
					       (remove val (gethash key tbl)))
					 tbl)
			    (message
			     "glossario: %s [%s] per %s in %s --da %s [%s]"
			     new (substring str (car new) (cadr new))
			     (nthcdr 3 (car val))
			     (substring x 0 -2)
			     old (and oldtxt (substring
					      oldtxt (car old) (cadr old)))))
		       (throw 'err nil)))))))))
     (mapcar #'car gi/roth/gls-rtables))))

(defun gi/roth/next (lne)
  (let ((go (gi/roth/following lne)))
    (while (gethash go gi/roth/trd-text)
      (setq go (gi/roth/following go)))
    (gi/roth/buffers go)))

(defun gi/roth/writeout ()
  (interactive)
  (let ((ctrd (mapcar (lambda (x) (symbol-value (intern-soft (car x))))
		      gi/roth/trd-tables))
	(cgls (mapcar (lambda (x) (symbol-value (intern-soft (car x))))
		      gi/roth/gls-tables))
	(new ()))
    (load gi/roth/traduzione) (load gi/roth/glossary)
    (cl-mapcar (lambda (x y) (maphash
			 (lambda (k v)
			   (when (not (equal v (gethash k (symbol-value
							   (intern-soft
							    (car y))))))
			     (push (cons (car y) (prin1-to-string k)) new)))
			 x))
	       (append cgls ctrd)
	       (append gi/roth/gls-tables gi/roth/trd-tables))
    (if (car new) (let ((max-mini-window-height (+ 1 (length new)))
			(msg (mapconcat
			      #'identity
			      (mapcar (lambda (x) (format " %s: %s"
						     (car x) (cdr x)))
				      new)
			      "\n")))
		    (when (y-or-n-p (concat "cambiamenti:\n" msg "? "))
		      (gi/roth/printt ctrd gi/roth/trd-tables gi/roth/traduzione)
		      (gi/roth/printt cgls gi/roth/gls-tables gi/roth/glossary)
		      (load gi/roth/traduzione) (load gi/roth/glossary)))
      (message "nessun cambiamento!"))))

;;; chords

(defvar gi/roth/keys)
(defun gi/roth/keys ()
  (let ((km (make-sparse-keymap)))
    (mapc
     (lambda (x) (define-key km (kbd (car x)) (cdr x)))
     '(("<left>"			; glossary left
	. (lambda () (interactive) (gi/roth/edit-glossary 'back)))
       ("<right>"			; glossary right
	. gi/roth/edit-glossary)
       ("<up>"				; glossary up
	. (lambda () (interactive) (gi/roth/choose-word 'up)))
       ("<down>"			; glossary down
	. gi/roth/choose-word)
       ("M-<left>"			; kill word
	. (lambda () (interactive) (let ((inhibit-read-only t)) (kill-word nil))))
       ("M-<right>"			; yank
	. (lambda () (interactive) (let ((inhibit-read-only t)) (yank))))
       ("M-<up>"			; transpose forward
	. (lambda () (interactive) (let ((inhibit-read-only t))
				(transpose-words -1))))
       ("M-<down>"			; transpose backward
	. (lambda () (interactive) (let ((inhibit-read-only t))
				(transpose-words 1))))
       ("C-<return>"			; choose translation
	. gi/roth/edit-word)
       ("M-<return>"			; new translation
	. (lambda () (interactive) (gi/roth/edit-word 'new)))
       ("C-c C-r"			; remove from glossary
	. gi/roth/rm-current-word)
       ("C-c C-n"			; next untranslated line
	. (lambda () (interactive) (gi/roth/next gi/roth/current-line)))
       ("C-c M-n"			; next line
	. (lambda () (interactive) (gi/roth/buffers
			       (gi/roth/following gi/roth/current-line))))
       ("C-c C-p"			; previous line
	. (lambda () (interactive) (gi/roth/buffers
			       (gi/roth/previous gi/roth/current-line))))
       ("C-c C-i"			; italicize region
	. gi/roth/edit-italic)
       ("C-c C-f"			; insert footnote ref
	. (lambda () (interactive) (gi/roth/edit-refs :note)))
       ("C-c C-l"			; insert link
	. (lambda () (interactive) (gi/roth/edit-refs :link)))
       ("C-c C-d"			; insert index ref
	. (lambda () (interactive) (gi/roth/edit-refs :index)))
       ("C-c C-a"			; nota di traduzione
	. gi/roth/ndt)
       ("C-c C-c"			; format line
	. (lambda () (interactive) (gi/roth/line-trd gi/roth/current-line)))
       ("C-c C-w"			; save changes
	. (lambda () (interactive) (gi/roth/writeout)))))
    (setq gi/roth/keys km)))

;;; startup

(defun gi/roth/startup ()
  (interactive)
  (let ((go '(0 0 0)))
    (gi/roth/load)
    (gi/roth/split)
    (gi/roth/next go)))

;;; index fixups

;;;; so so ugly...
(defvar gi/roth/fix-indexref ())
(defun gi/roth/fix-indexref () ; (gi/roth/fix-indexref)
  (let ((a ()) (b ()) (c ()) (l '(19 0 0)) (e ()) (z ()))
    (with-temp-buffer
      (insert-file-contents (concat gi/roth/basedir "xhtml/" "back02.xhtml"))
      (setq a (cl-remove-if-not (lambda (x) (and (listp x) (equal (car x) 'p)))
				(cdr (assoc 'body
					    (libxml-parse-html-region
					     (point-min) (point-max)))))))
    (dolist (x a) (dolist (y x) (when (and (listp y) (equal (car y) 'a))
				  (push y z)))
	    (push z b) (setq z ()))
    (dolist (x b) (dolist (y x) (let ((lnk (cdar (nth 1 y)))
				      (lab (nth 2 y)))
				  (push (cons lnk lab) z)))
	    (push z c) (setq z ()))
    (setq d (copy-sequence c))
    (while d
      (while l (when (or (equal "voce indice"
				(gethash (butlast l) gi/roth/elt-type))
			 (equal "sottovoce indice"
				(gethash (butlast l) gi/roth/elt-type)))
		 (let ((lab (car d))
		       (ref (gethash l gi/roth/lne-indexref)))
		   (when (and (equal (length lab) (length ref))
			      (cl-every (lambda (x y) (equal (car x) (cdr y))) lab ref))
		     (push (cons l (pop d)) e))))
	     (setq l (gi/roth/following l)))
      (setq d (cdr d) l (gi/roth/following (caar e))))
    (setq gi/roth/fix-indexrefs e)))

(defun gi/roth/fixed-indexref (lne str)
  (or (cdr (assoc str (cdr (assoc lne gi/roth/fix-indexrefs))))
      (progn (string-match "#p\\(?1:[[:digit:]]+$\\)" str)
	     (match-string 1 str))))

(defvar gi/roth/indice ())
(defun gi/roth/indice () ; (gi/roth/indice)
  (let ((l '(19 1 0)) (e '(20 0 0))
	(a ()) (b1 ()) (b2 ()) (b3 ()) (c ()))
    (while (not (equal l e))
      (cond
       ((cl-some (lambda (x) (equal l x))
		 '((19 363 0) (19 363 1) ; job/labor market
		   (19 392 0) ; national
		   (19 503 0) (19 518 0) (19 769 0) ; acronimi ridondanti
		   )))
       ((cl-some (lambda (x) (equal l x)) '((19 326 0) (19 327 0) (19 328 0)
				       (19 329 0) (19 330 0) (19 331 0)))
	(push l b3))
       ((equal "sottovoce indice" (gethash (butlast l) gi/roth/elt-type))
	(progn (push l c)
	       (when (not (equal (nth 1 (gi/roth/following l)) (nth 1 l)))
		 (setcdr (assoc (caar b1) b2)
			 (cons (sort c #'gi/roth/key<=)
			       (cdr (assoc (caar b1) b2))))
		 (setq c ()))))
       (t (progn (push l c)
		 (when (not (equal (nth 1 (gi/roth/following l)) (nth 1 l)))
		   (push (sort c #'gi/roth/key<=) b1)
		   (push (cons (caar b1) nil) b2)
		   (setq c ())))))
      (setq l (gi/roth/following l)))
    (setq b2 (cl-remove-if #'null (mapcar (lambda (x) (when (> (length x) 1) x))
					  b2))
	  b3 (list '(19 235 0) b3))
    (cl-flet ((lx (lambda (x y) (string-collate-lessp
			    (gethash (car x) gi/roth/trd-text)
			    (gethash (car y) gi/roth/trd-text)
			    "it_IT.UTF-8" t)))) 
      (setq b1 (sort b1 #'lx))
      (dolist (x b1)
	(dolist (y x)
	  (push y a))
	(when (car (assoc (car x) b2))
	  (dolist (y (sort (cdr (assoc (car x) b2)) #'lx))
	    (dolist (z y)
	      (push z a))
	    (when (car (assoc (car y) b3))
	      (dolist (w (sort (cdr (assoc (car y) b3)) #'lx))
		(push w a)))))))
    (setq gi/roth/indice (reverse a))))
